# -*- coding: utf-8 -*-
from loggingd import log_enter
from textminer import util
import doctest
import json
import yaml

def compile(rule, fmt='yaml'):
    '''
    >>> rule = {'value': {'prefix': '<body>', 'suffix': '</body>'}}
    >>> parser = compile(rule, fmt=None)
    >>> type(parser).__name__
    'ValueExtractor'
    '''
    from textminer import extractors
    rule = _parse_rule(rule, fmt)
    keys = list(rule.keys())
    if len(keys) != 1:
        raise Exception('Rule should have one single root element.')
    parser_type = keys[0]
    parser_rule = rule[parser_type]
    return extractors.create(parser_type, parser_rule)
    
def extract(text, rule, fmt='yaml'):
    '''
    >>> extract('<html><body>abc</body></html>', {'value': {'prefix': '<body>', 'suffix': '</body>'}}, fmt=None)
    'abc'
    '''
    extractor = compile(rule, fmt=fmt)
    return extractor.extract(text)

@log_enter('[DEBUG] Crawling {url} ...')
def extract_from_url(url, rule, fmt='yaml', charset=None):
    html = util.curl(url, charset=charset)
    return extract(html, rule, fmt=fmt)

def _parse_rule(rule, fmt):
    if fmt is None:
        return rule
    elif fmt == 'yaml':
        return yaml.load(rule)
    elif fmt == 'json':
        return json.loads(rule)
    else:
        raise Exception('Unknown rule format "%s".' % fmt)

if __name__ == '__main__':
    doctest.testmod()
    


# -*- coding: utf-8 -*-
from datetime import datetime as DateTime
from textminer import util
import doctest
import re
import six

def default(value, default_value, empty_value=None):
    '''
    >>> default('1', '0')
    '1'
    >>> default(None, '0')
    '0'
    >>> default('', '0')
    ''
    >>> default('', '0', empty_value='')
    '0'
    '''
    return default_value if value == empty_value else value

def date(value, format='%Y-%m-%d'):
    '''
    >>> date('2000-01-01')
    datetime.date(2000, 1, 1)
    '''
    return DateTime.strptime(value, format).date()

def datetime(value, format='%Y-%m-%d %H:%M:%S'):
    '''
    >>> datetime('2000-01-01 12:30:40')
    datetime.datetime(2000, 1, 1, 12, 30, 40)
    '''
    return DateTime.strptime(value, format)

def eval(value, expression):
    '''
    >>> eval(10, 'value / 10') == 1
    True
    '''
    if six.PY2:
        import __builtin__ as builtins
    else:
        import builtins
    return builtins.eval(expression, {}, {'value': value})

def strip(value, chars=None):
    '''
    >>> strip(' \tabc def\t ')
    'abc def'
    >>> strip('"abc"', '"')
    'abc'
    '''
    return value.strip(chars)

def strip_html(value):
    '''
    >>> strip_html('a<b>b</b>c')
    'abc'
    '''
    value = util.compact_html(value)
    return re.sub('<.+?>', '', value)

def _none_wrapper(func):
    '''
    >>> def func(value): return value + 1
    >>> func = _none_wrapper(func)
    >>> func(1)
    2
    >>> func(None) is None
    True
    '''
    def wrapper(value, *args, **kw):
        if value is None:
            return None
        else:
            return func(value, *args, **kw)
    return wrapper

def _number_wrapper(func):
    '''
    >>> func = _number_wrapper(int)
    >>> func('1,234,567')
    1234567
    '''
    def wrapper(value, *args, **kw):
        value = value.replace(',', '')
        return func(value, *args, **kw)
    return wrapper

FILTERS = {
    'default': default,
    'bool': _none_wrapper(bool),
    'date': _none_wrapper(date),
    'datetime': _none_wrapper(datetime),
    'eval': _none_wrapper(eval),
    'float': _none_wrapper(_number_wrapper(float)),
    'int': _none_wrapper(_number_wrapper(int)),
    'strip': _none_wrapper(strip),
    'strip_html': _none_wrapper(strip_html),
}
    
if __name__ == '__main__':
    doctest.testmod()
    


# -*- coding: utf-8 -*-
import doctest
import re

class Matcher(object):
    def match(self, text, start):
        raise NotImplementedError()
    
class StringMatcher(Matcher):
    def __init__(self, pattern):
        super(StringMatcher, self).__init__()
        self._pattern = pattern
        
    def match(self, text, start):
        '''
        >>> m = StringMatcher('1')
        >>> m.match('a1a1a', 0)
        ('1', 1, 2)
        >>> m.match('a1a1a', 2)
        ('1', 3, 4)
        >>> m.match('a2a', 0)
        (None, -1, -1)
        '''
        index = text.find(self._pattern, start)
        if index != -1:
            return self._pattern, index, index + len(self._pattern)
        else:
            return None, -1, -1
        
class RegexMatcher(Matcher):
    def __init__(self, pattern):
        super(RegexMatcher, self).__init__()
        self._pattern = re.compile(pattern)
        
    def match(self, text, start):
        '''
        >>> m = RegexMatcher('\\d')
        >>> m.match('a1a1a', 0)
        ('1', 1, 2)
        >>> m.match('a1a1a', 2)
        ('1', 3, 4)
        >>> m.match('aba', 0)
        (None, -1, -1)
        '''
        match = self._pattern.search(text[start:])
        if match:
            return match.group(), start + match.start(), start + match.end()
        else:
            return None, -1, -1
    
if __name__ == '__main__':
    doctest.testmod()
    


# -*- coding: utf-8 -*-
import doctest
import re
import requests

class Dict(dict):
    '''
    A Dict object is like a dictionary except `obj.foo` can be used
    in addition to `obj['foo']`.
    
    >>> o = Dict(a=1)
    >>> o.a
    1
    >>> o['a']
    1
    >>> o.a = 2
    >>> o['a']
    2
    >>> del o.a
    >>> o.a
    Traceback (most recent call last):
    ...
    AttributeError: 'a'
    >>> del o.b
    Traceback (most recent call last):
    ...
    AttributeError: 'b'
    '''
    def __getattr__(self, key):
        try:
            return self[key]
        except KeyError as e:
            raise AttributeError(str(e))
    
    def __setattr__(self, key, value): 
        self[key] = value
    
    def __delattr__(self, key):
        try:
            del self[key]
        except KeyError as e:
            raise AttributeError(str(e))
        
def compact_html(html):
    '''
    >>> compact_html('<div id="div1">\\n abc def \\n</div>')
    '<div id="div1">abc def</div>'
    '''
    html = re.sub('\\s', ' ', html)
    html = re.sub(' +', ' ', html)
    return html.replace(' <', '<').replace('> ', '>')

def curl(url, charset=None):
    headers = {'User-Agent': 'Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Ubuntu Chromium/43.0.2357.81 Chrome/43.0.2357.81 Safari/537.36'}
    resp = requests.get(url, headers=headers, timeout=30)
    resp.raise_for_status()
    if resp.headers.get('Content-Type', '').startswith('text/html'):
        charset = _detect_charset(resp.text[:1024])
        if charset is not None:
            resp.encoding = charset
    return compact_html(resp.text)

_CHARSET_PATTERNS = [
    re.compile('<meta http-equiv="Content-Type" content="text/html; charset=(.*)"', re.IGNORECASE),
    re.compile('<meta charset="(.*)"', re.IGNORECASE),
]
def _detect_charset(head):
    '''
    >>> html = '<html><head><meta http-equiv="Content-Type" content="text/html; charset=utf-8" /></head>'
    >>> _detect_charset(html)
    'utf-8'
    
    >>> html = '<html><head><meta http-equiv="Content-Type" content="text/html; charset=utf-8"></head>'
    >>> _detect_charset(html)
    'utf-8'
    
    >>> html = '<html><head><meta http-equiv="Content-Type" content="text/html; charset=utf-8" /></head>'
    >>> _detect_charset(html)
    'utf-8'
    
    >>> html = '<html><head><meta charset="utf-8"></head>'
    >>> _detect_charset(html)
    'utf-8'
    
    >>> html = '<html>'
    >>> _detect_charset(html) is None
    True
    '''
    for pattern in _CHARSET_PATTERNS:
        match = pattern.search(head)
        if match is not None:
            return match.group(1)
        
if __name__ == '__main__':
    doctest.testmod()
    


# -*- coding: utf-8 -*-
from textminer import filters
from textminer.matchers import RegexMatcher, StringMatcher
from textminer.util import Dict
import doctest
import loggingd

log = loggingd.getLogger(__name__)

class ExtractError(Exception):
    pass

class Extractor(object):
    def __init__(self, rule):
        self._rule = rule
        self._prefix_matcher = None
        self._suffix_matcher = None
        self._filters = None
        self._child = None
        self._compile(rule)
        
    def extract(self, text):
        try:
            return self._extract(text)
        except:
            short_text = text[:100] + ' ...' if len(text) > 100 else text
            msg = 'Failed to extract.\nRule:\n%s\nText:\n%s' % (self._rule, short_text)
            log.warn(msg, exc_info=True)
            raise ExtractError(msg)
        
    def _extract(self, text):
        raise NotImplementedError()
    
    def _compile(self, rule):
        self._prefix_matcher = _compile_pattern(rule['prefix'])
        self._suffix_matcher = _compile_pattern(rule['suffix'])
        self._filters = _compile_filters(rule.get('filters'))
        self._child = _compile_child(rule)
        
    def _filter(self, value):
        '''
        >>> p = Extractor({'prefix': '<html>', 'suffix': '</html>'})
        >>> p._filters = ['default(value, "0")', 'int(value)']
        >>> p._filter(None)
        0
        '''
        for f in self._filters:
            value = eval(f, filters.FILTERS, {'value': value})
        return value
        
    def _process_child(self, value):
        '''
        >>> e = Extractor({'prefix': '<html>', 'suffix': '</html>'})
        >>> e._process_child('<div>aaa</div>')
        '<div>aaa</div>'
        
        >>> e._child = ValueExtractor({'prefix': '<div>', 'suffix': '</div>'})
        >>> e._process_child('<div>aaa</div>')
        'aaa'
        
        >>> e._process_child(None) is None
        True
        '''
        if self._child is None or value is None:
            return value
        return self._child.extract(value)
        
class ValueExtractor(Extractor):
    def _extract(self, text):
        value, _ = self.parse_with_end(text)
        return value
    
    def parse_with_end(self, text):
        '''
        >>> p = ValueExtractor({'prefix': '<body>', 'suffix': '</body>'})
        >>> p.parse_with_end('<html><body>abc</body></html>')
        ('abc', 22)
        >>> p.parse_with_end('<html><header>abc</header></html>')
        (None, 0)
        '''
        value, end = _search(self._prefix_matcher, self._suffix_matcher, text)
        value = self._filter(value)
        value = self._process_child(value)
        return value, end
    
class ListExtractor(Extractor):
    def _extract(self, text):
        '''
        >>> p = ListExtractor({'prefix': '<li>', 'suffix': '</li>'})
        >>> p.extract('<ul><li>item1</li><li>item2</li></ul>')
        ['item1', 'item2']
        >>> p.extract('<ul></ul>')
        []
        '''
        items = []
        end = 0
        while True:
            value, end = _search(self._prefix_matcher, self._suffix_matcher, text, end)
            if value is None:
                break
            value = self._filter(value)
            value = self._process_child(value)
            items.append(value)
        return items
    
class DictExtractor(Extractor):
    def __init__(self, *args, **kw):
        self._keys = None
        self._children = None
        super(DictExtractor, self).__init__(*args, **kw)
        
    def _compile(self, rule):
        self._keys = [field['key'] for field in rule]
        self._children = [create('value', field) for field in rule]
        
    def _extract(self, text):
        '''
        >>> p = DictExtractor([{'key': 'name', 'prefix': '<td id="1">', 'suffix': '</td>'}, {'key': 'value', 'prefix': '<td id="2">', 'suffix': '</td>'}])
        
        >>> d = p.extract('<tr><td id="1">abc</td><td id="2">123</td></tr>')
        >>> d['name']
        'abc'
        >>> d['value']
        '123'
        
        >>> d = p.extract('<tr></tr>')
        >>> d['name'] is None
        True
        >>> d['value'] is None
        True
        '''
        values = Dict()
        end = 0
        for i in range(len(self._keys)):
            values[self._keys[i]], temp_end = self._children[i].parse_with_end(text[end:])
            end += temp_end
        return values
    
def create(parser_type, rule):
    '''
    >>> rule = {'prefix': '<body>', 'suffix': '</body>'}
    >>> parser = create('value', rule)
    >>> type(parser).__name__
    'ValueExtractor'
    '''
    parser_class = globals().get(parser_type.capitalize() + 'Extractor')
    if parser_class is None:
        raise Exception('Extractor with type="%s" could not be found.' % parser_type)
    return parser_class(rule)

def _compile_child(rule):
    '''
    >>> rule = {'prefix': '<body>', 'suffix': '</body>'}
    >>> _compile_child(rule) is None
    True
    
    >>> rule = {'prefix': '<body>', 'suffix': '</body>', 'value': {'prefix': '<b>', 'suffix': '</b>'}}
    >>> child = _compile_child(rule)
    >>> type(child).__name__
    'ValueExtractor'
    '''
    for parser_type in ['value', 'list', 'dict']:
        if parser_type in rule:
            return create(parser_type, rule[parser_type])
    else:
        return None

def _compile_filters(filters):
    '''
    >>> _compile_filters(['int'])
    ['int(value)']
    >>> _compile_filters(['default(0)'])
    ['default(value, 0)']
    >>> _compile_filters(None)
    []
    '''
    results = []
    if filters is not None:
        for f in filters:
            if '(' in f:
                f = f.replace('(', '(value, ', 1)
            else:
                f = f + '(value)'
            results.append(f)
    return results

def _compile_pattern(pattern):
    '''
    normal text
    >>> _compile_pattern('a|b').match('a|b', 0)
    ('a|b', 0, 3)
    >>> _compile_pattern('//a|b//').match('/a|b/', 0)
    ('a|b', 1, 4)
    
    regular expression
    >>> _compile_pattern('/a|b/').match('b', 0)
    ('b', 0, 1)
    
    single slash
    >>> _compile_pattern('/').match('a/b', 0)
    ('/', 1, 2)
    '''
    if pattern.startswith('//') and pattern.endswith('//'):
        return StringMatcher(pattern[2:-2])
    elif len(pattern) >= 2 and pattern.startswith('/') and pattern.endswith('/'):
        return RegexMatcher(pattern[1:-1])
    else:
        return StringMatcher(pattern)
    
def _search(prefix_matcher, suffix_matcher, text, start_index=0):
    '''
    >>> prefix_matcher = StringMatcher('<b>')
    >>> suffix_matcher = StringMatcher('</b>')
    
    prefix & suffix found
    >>> _search(prefix_matcher, suffix_matcher, '<div><b>abc</b></div>', 1)
    ('abc', 15)
    
    prefix not found
    >>> _search(prefix_matcher, suffix_matcher, '<div><span>abc</span></div>', 1)
    (None, 1)
    
    prefix found but suffix not found
    >>> _search(prefix_matcher, suffix_matcher, '<div><b>abc</span></div>', 1)
    (None, 1)
    
    prefix is start and suffix is end
    >>> _search(RegexMatcher('^'), RegexMatcher('$'), 'abc')
    ('abc', 3)
    '''
    # find prefix
    matched, _, match_end = prefix_matcher.match(text, start_index)
    if matched is None:
        return (None, start_index)
    start = match_end
    
    # find suffix
    matched, match_start, match_end = suffix_matcher.match(text, start)
    if matched is None:
        return (None, start_index)
    end = match_end
    
    # extract text
    return text[start:match_start], end

if __name__ == '__main__':
    doctest.testmod()
    